import traceback
from typing import List

try:
    # Import third party libs
    from starlette.applications import Starlette
    from starlette.responses import JSONResponse
    from starlette.routing import Route
    from starlette.requests import Request
    from uvicorn import Server
    from uvicorn.config import Config

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __init__(hub):
    hub.pop.sub.add(dyne_name="rend")
    hub.srv.starlette.SERVER = None


def __virtual__(hub):
    return HAS_LIBS


async def start(
    hub,
    host: str,
    port: int,
    matcher_plugin: str,
    prefix: str,
    refs: List[str],
    limit_concurrency: int = None,
):
    async def _query(request):
        return await hub.srv.starlette.query(
            request, prefix=prefix, refs=refs, matcher_plugin=matcher_plugin
        )

    routes = [
        Route("/", _query, methods=["GET", "POST", "PUT"]),
    ]
    app = Starlette(debug=True, routes=routes)
    config = Config(
        app,
        host=host,
        port=port,
        loop="none",
        limit_concurrency=limit_concurrency,
        log_level=hub.log.INT_LEVEL,
    )
    hub.srv.starlette.SERVER = Server(config)

    return await hub.srv.starlette.SERVER.serve()


async def join(hub):
    """
    Block until the server has started
    """
    while not hasattr(hub.srv.starlette, "SERVER"):
        await hub.pop.loop.sleep(0)

    while not getattr(hub.srv.starlette.SERVER, "started", False):
        await hub.pop.loop.sleep(0)

    # Wait just one more cycle for everything to start
    await hub.pop.loop.sleep(0)


async def stop(hub):
    if hub.srv.starlette.SERVER:
        hub.srv.starlette.SERVER.should_exit = True
        hub.srv.starlette.SERVER.force_exit = True
        await hub.srv.starlette.SERVER.shutdown()


async def query(
    hub, request: Request, matcher_plugin: str, prefix: str, refs: List[str]
) -> JSONResponse:
    if request.headers["Content-type"] == "application/json":
        q_params = await request.json()
    else:
        q_params = dict(request.query_params)

    if "ref" not in q_params:
        return JSONResponse(
            {"error": "Required ref not found in params"}, status_code=417
        )

    q_ref = q_params.get("ref")
    if prefix:
        q_ref = f"{prefix}.{q_ref}"

    if not hub.matcher[matcher_plugin].match(q_ref, prefix=prefix, refs=refs):
        return JSONResponse(
            {"error": f"The provided ref {q_ref} is not available for execution"},
            status_code=405,
        )

    if not hasattr(hub, q_ref):
        return JSONResponse(
            {"error": f"The provided ref {q_ref} does not exist"}, status_code=404
        )

    try:
        ret = await hub.srv.init.runner(
            q_ref,
            *q_params.get("args", []),
            **q_params.get("kwargs", {}),
        )
        return JSONResponse(ret, status_code=200)
    except Exception as e:
        hub.log.error(traceback.format_exc())
        return JSONResponse(
            {"error": f"hub.{q_ref} returned {e.__class__.__name__}: {e}"},
            status_code=500,
        )



import fnmatch
from typing import List


def match(hub, ref: str, prefix: str, refs: List[str]) -> bool:
    """
    Check the desired ref to see if it matches a globular ref in the refs list
    """
    for pattern in refs:
        if prefix:
            pattern = f"{prefix}.{pattern}"
        if fnmatch.fnmatch(ref, pattern):
            return True
    else:
        return False



CLI_CONFIG = {
    "host": {"group": "Gate"},
    "port": {"group": "Gate"},
    "server": {"group": "Gate"},
    "prefix": {"group": "Gate"},
    "matcher": {"group": "Gate"},
    "refs": {"group": "Gate"},
    "limit_concurrency": {"group": "Gate"},
}
CONFIG = {
    "host": {
        "default": "0.0.0.0",
        "type": str,
        "help": "The host interface to bind to.",
    },
    "port": {"default": 8080, "type": int, "help": "The port to bind to"},
    "limit_concurrency": {
        "default": None,
        "type": int,
        "help": "The Maximum number of concurrent connections or tasks to allow, before issuing HTTP 503 responses.  "
        "This is useful for ensuring known memory usage patterns even under over-resourced loads.",
    },
    "server": {"default": "starlette", "help": "The server plugin interface to use"},
    "prefix": {
        "default": None,
        "type": str,
        "help": "The prefix for hub references that can be exposed",
    },
    "matcher": {
        "default": "glob",
        "type": str,
        "help": "The matcher plugin that will be used on the refs",
    },
    "refs": {
        "default": ["gate.init.test"],
        "help": "The hub references to expose",
        "nargs": "*",
    },
}
SUBCOMMANDS = {}
DYNE = {
    "gate": ["gate"],
    "matcher": ["matcher"],
    "srv": ["srv"],
}



import json
from typing import Any
from typing import Dict
from typing import List


def __init__(hub):
    hub.pop.sub.add(dyne_name="tree")
    hub.pop.sub.add(dyne_name="matcher")
    hub.pop.sub.add(dyne_name="srv")


def cli(hub):
    hub.pop.config.load(["gate"], cli="gate")
    hub.pop.loop.create()
    coro = hub.gate.init.start(
        gate_server=hub.OPT.gate.server,
        host=hub.OPT.gate.host,
        port=hub.OPT.gate.port,
        matcher_plugin=hub.OPT.gate.matcher,
        prefix=hub.OPT.gate.prefix,
        refs=hub.OPT.gate.refs,
        limit_concurrency=hub.OPT.gate.limit_concurrency,
    )
    hub.pop.Loop.run_until_complete(coro)


async def start(
    hub,
    gate_server: str,
    host: str,
    port: int,
    matcher_plugin: str,
    prefix: str,
    refs: List[str],
    limit_concurrency: int = None,
):
    return await hub.srv[gate_server].start(
        host=host,
        port=port,
        matcher_plugin=matcher_plugin,
        prefix=prefix,
        refs=refs,
        limit_concurrency=limit_concurrency,
    )


async def tree(
    hub, prefix: str = None, refs: List[str] = None, matcher_plugin: str = None
) -> Dict[str, Any]:
    """
    Get details of all references that are exposed through the gate API
    """
    if prefix is None:
        prefix = hub.OPT.gate.prefix
    if matcher_plugin is None:
        matcher_plugin = hub.OPT.gate.matcher
    if refs is None:
        refs = hub.OPT.gate.refs

    _tree: Dict[str, Any] = await hub.tree.init.traverse()

    def _get_ref(t: Dict[str, Any]):
        rf = t.get("ref")
        if (
            rf
            and isinstance(rf, str)
            and hub.matcher[matcher_plugin].match(t["ref"], prefix=prefix, refs=refs)
        ):
            # This item has a ref and is available to the api
            return t
        else:
            r = {}
            for k, v in t.items():
                if isinstance(v, Dict):
                    sub_ref = _get_ref(v)
                    if sub_ref:
                        r[k] = sub_ref
            return r

    ret = _get_ref(_tree)

    return json.loads(hub.output.json.display(ret))


async def stop(hub, gate_server: str):
    hub.log.debug("Shutting down gate server")
    return await hub.srv[gate_server].stop()


async def join(hub, gate_server):
    return await hub.srv[gate_server].join()


async def test(hub, *args, **kwargs):
    """
    The test function for gate
    """
    return {"args": args, "kwargs": kwargs}



from typing import Any


def __init__(hub):
    # A mapping of dyne_names/subs to a reference to a function on the hub that should run that function
    # If no reference is listed then a getattr(hub, ref) will be done and the reference will be called directly
    hub.srv.RUNNER = {}


async def runner(hub, ref: str, *args, **kwargs) -> Any:
    """
    Call a function based on runners defined before the program started
    """
    new_args = [x for x in args]
    prefix = ref.split(".")[0]
    if prefix in hub.srv.RUNNER:
        # pass the ref to the function
        new_args.insert(1, ref)
        func = hub.srv.RUNNER[prefix]
        if not hasattr(func, "__call__"):
            # If the prefix was a string then get it's reference on the hub
            func = hub[func]
    else:
        func = hub[ref]

    if new_args and kwargs:
        ret = hub.pop.loop.wrap(func, *new_args, **kwargs)
    elif new_args:
        ret = hub.pop.loop.wrap(func, *new_args)
    elif kwargs:
        ret = hub.pop.loop.wrap(func, **kwargs)
    else:
        ret = hub.pop.loop.wrap(func)

    ret = await hub.pop.loop.unwrap(ret)
    return ret
